#pragma once

#include <iostream>
#include <iomanip>

using namespace std;

namespace ADSLibrary
{
	namespace DataStructures
	{
		namespace ArrayBasedStructures
		{
			namespace Procedural
			{
				/**
				 * Maximln velikost seznamu.
				 */
				const int ListSize = 5;

				/**
				 * Ukazatel neukazujc na dnou poloku.
				 */
				const int Nil = -1;

				/**
				 * \brief Struktura reprezentujc poloku v seznamu {@link #ADSLibrary::DataStructures::ArrayBasedStructures::Procedural::List}
				 */
				struct ListItem
				{
					/**
					 * Hodnota obsaen v poloce.
					 */
					int Value;
					/**
					 * Ukazatel na pedchoz poloku. Pokud takov poloka neexistuje ukazatel m hodnotu {@link #Nil}.
					 */
					int Prev;
					/**
					 * Ukazatel na nsledujc  poloku. Pokud takov poloka neexistuje ukazatel m hodnotu {@link #Nil}.
					 */
					int Next;
				};

				/**
				 * \brief Obousmrn seznam implementovan v poli procedurlnm zpsobem.
				 *
				 * Implementace obousmrnho seznamu v poli. V seznamu je implementovn garbage collector a jednoduch itertor.
				 * Implementace v poli znamen, e ukazatel je v tomto ppad realizovn jako slo typu int a ukazatel je de-facto index v poli.
				 *
				 * @author	Ji Dvorsk <jiri.dvorsky@vsb.cz>
				 * @date	2010 - 2015
				 */
				struct List
				{
					/**
					 * Pole obsahujc data uloen do seznamu.
					 */
					ListItem Data[ListSize];
					/**
					 * Hlava seznamu tj. ukazatel na prvn poloku seznamu v poli {@link #Data}.
					 */
					int Head;
					/**
					 * Ukazatel na prvn volnou poloku v poli {@link #Data}. Pokud takov poloka neexistuje, m ukazatel hodnotu Nil.
					 * Tento ukazatel je vyuvn garbage collectorem.
					 */
					int Free;
					/**
					 * Ukazatel na aktuln poloku pi iteraci seznamem.
					 */
					int Current;
				};

				/**
				 * Inicializace seznamu.
				 * Protoe je seznam implementovan v poli, nemusme provdt dnou alokaci pamti a tud meme
				 * inicializaci seznamu povaovat za vpodstat smazn pvodnho obsahu.
				 * Pokud bychom pouvali dynamicky alokovan pamov struktury, museli bychom nejprve provst ppadnou
				 * alokaci potebn pamti.
				 *
				 * @param L Inicializovan seznam
				 */
				void Init(List& L);

				/**
				 * Smazn celho obsahu seznamu. Smazn spov v "zapomenut" struktury vzjemnch odkaz prvk seznamu mezi sebou.
				 *
				 * @param L Seznam
				 */
				void Clear(List& L);

				/**
				 * Vloen prvku do seznamu. Prvek je vloen ped souasnou hlavu seznamu tj. stane se novou hlavou seznamu.
				 *
				 * @param L Seznam
				 * @param X Vkldan prvek
				 */
				void Insert(List& L, const int X);

				/**
				 * Odebrn prvku ze seznamu. Prvek X je nejprve v seznamu nalezen a pot je odpovdajc poloka ze seznamu vymazna.
				 *
				 * @param L Seznam
				 * @param X Odebran prvek
				 */
				void Delete(List& L, const int X);

				/**
				 * Hledn prvku v seznamu.
				 *
				 * @param L Seznam
				 * @param X Hledan prvek
				 * @return Funkce vrac true v ppad, e je prvek X nalezen, jinak vrac false.
				 */
				bool Search(const List& L, const int X);

				/**
				 * Test je-li seznam przdn.
				 *
				 * @param L Seznam
				 * @return Funkce vrac true pokud je seznam przdn, jinak false.
				 */
				bool IsEmpty(const List& L);

				/**
				 * Funkce vrac poet prvk v seznamu.
				 *
				 * @param L Seznam
				 * @return Poet prvk v seznamu.
				 */
				int Count(const List& L);

				/**
				 * Reset itertoru. Ukazatel na aktuln poloku seznamu bude ukazovat na hlavu seznamu.
				 *
				 * @param L Seznam
				 */
				void Reset(List& L);

				/**
				 * Test ukonen innosti itertoru.
				 *
				 * @param L Seznam
				 * @return Funkce vrac true pokud je iterace seznamem ukonena, jinak false.
				 */
				bool IsEnd(const List& L);

				/**
				 * Posun ukazatele aktuln poloky na dal poloku v seznamu.
				 *
				 * @param L Seznam
				 */
				void MoveNext(List& L);

				/**
				 * Vrcen prvku z aktuln poloky seznamu pi iteraci.
				 *
				 * @param L Seznam
				 * @return Funkce vrac prvek z aktuln poloky seznamu pi iteraci.
				 */
				int Current(const List& L);

				/**
				 * Alokace voln poloky v seznamu. Alokovan poloka se odebrna z prvn pozice v garbage collectoru.
				 *
				 * @param L Seznam
				 * @return Funkce vrac ukazatel na alokovanou poloku v seznamu. V ppad, e ji nejsou dn voln poloky k dispozici, funkce vrac Nil.
				 */
				int AllocateItem(List& L);

				/**
				 * Dealokace, uvolnn, poloky v seznamu.
				 *
				 * @param L Seznam
				 * @param Index Ukazatel na uvolovanou poloku v seznamu. Uvolnn poloka je vrcena na prvn msto v garbage collecotoru.
				 */
				void FreeItem(List& L, int Index);

				/**
				 * Pomocn funkce, kter zobrazuje vnitn strukturu seznamu.
				 *
				 * @param L Seznam
				 */
				void Report(const List& L);
			}
		}
	}
}
