#pragma once

#include <iostream>
#include <string>

using namespace std;

namespace ADSLibrary
{
	namespace DataStructures
	{
		namespace LinkedStructures
		{
			namespace OOP
			{
				/**
	 * Implementace binrnho vyhledvacho stromu.
	 * Strom je implementovn jako nevyven, tud me zdegenerovat do seznamu.
	 * Ve stromu je maximln vyuito rekurze, veejn metody tvo vtinou jen "oblku" pro voln privtnch rekurzvnch metod.
	 * Tyto veejn "oblky" jsou proto implementovny jako inline.
	 *
	 * @author      Ji Dvorsk <jiri.dvorsky@vsb.cz>
	 */
				class BinarySearchTree
				{
				public:
					/**
					 * Konstruktor.
					 */
					BinarySearchTree();

					/**
					 * Destruktor.
					 */
					~BinarySearchTree();

					/**
					 * Vloen kle do stromu.
					 * @param Key Vkldan kl
					 */
					void Insert(const string Key);

					/**
					 * Odebrn kle ze stromu.
					 * @param Key Odebran kl
					 */
					void Delete(const string Key);

					/**
					 * Hledn kle ve stromu s vyuitm rekurze.
					 * @param Key Hledan kl
					 * @return Funkce vrac true v ppad, e je prvek Key nalezen, jinak vrac false.
					 */
					bool Search(const string Key) const;

					/**
					 * Hledn prvku ve stromu s vyuitm cyklu.
					 * @param Key Hledan kl
					 * @return Funkce vrac true v ppad, e je kl Key nalezen, jinak vrac false.
					 */
					bool SearchIterative(const string Key) const;

					/**
					 * Smazn celho stromu.
					 */
					void Clear();

					/**
					 * Test je-li strom przdn.
					 * @return Funkce vrac true pokud je seznam przdn, jinak false.
					 */
					bool IsEmpty() const;

					/**
					 * Funkce vrac poet uzl ve stromu.
					 * @return Poet uzl ve stromu.
					 */
					int Count() const;

					/**
					 * Prchod stromem zpsobem in order. Metoda vol privtn demostran funkci DoSomeAction.
					 */
					void InorderTraversal();

					/**
					 * Prchod stromem zpsobem pre order. Metoda vol privtn demostran funkci DoSomeAction.
					 */
					void PreorderTraversal();

					/**
					 * Prchod stromem zpsobem post order. Metoda vol privtn demostran funkci DoSomeAction.
					 */
					void PostorderTraversal();

					/**
					 * Vpis intern struktury stromu.
					 * Strom je vypisovn po "patrech", vdy je vypsn nejdve koen stromu, pak cel jeho lev podstrom a pak prav podstrom.
					 * Jednotliv "patra" jsou odsazovna.
					 */
					void ReportStructure() const;


					/**
					 * Deklarace sptelench td, kter implementuj itertory.
					 * V tomto prklad jsou itertory implementovny oddlen, tud lze pouvat vce itertor pro jeden strom souasn. Veker data nutn pro iterovn skrz strom si uchovv itertor.
					 * Deklarace tchto td jako sptelench td znamen, e tyto tdy maj pstup k privtnm lenm tdy BinarySearchTree, ale souasn nen nutn z tchto privtnch len nic zveejovat.
					 */
					friend class BinaryTreeIteratorBase;
					friend class BinaryTreeIterator;
					friend class BinaryTreeReverseIterator;

				private:
					/**
					 * Struktura reprezentujc uzel ve stromu.
					 */
					struct Node
					{
						/**
						 * Kl obsaen v uzlu.
						 */
						string Key;
						/**
						 * Ukazatel na lev podstrom.
						 */
						Node* Left;
						/**
						 * Ukazatel na prav podstrom.
						 */
						Node* Right;
					};

					/**
					 * Rekurzvn metoda pro vkldn kle do stromu.
					 * @param p Aktuln uzel stromu.
					 * @param Key Vkldan kl.
					 */
					void InternalInsert(Node*& p, const string Key);

					/**
					 * Rekurzvn metoda pro smazn kle ze stromu.
					 * @param p Aktuln uzel stromu.
					 * @param Key Odebran kl.
					 */
					void InternalDelete(Node*& p, const string Key);

					/**
					 * Rekurzvn metoda pro smazn kle ze stromu, uzel q m dva potomky, hledme nejpravjho potomka z levho podstromu uzlu q.
					 * Jinak eeno hledme nejbli men kl.
					 * @param p Aktuln uzel stromu.
					 * @param q Uzel kter se bude ruit.
					 */
					void InternalDeleteRightmostFromLeftSubtree(Node*& r, Node*& q);

					/**
					 * Rekurzvn metoda pro hledn kle ze stromu.
					 * @param p Aktuln uzel stromu.
					 * @param Key Hledan kl.
					 */
					bool InternalSearch(const Node* p, const string Key) const;

					/**
					 * Rekurzvn metoda pro postupn smazn celho stromu.
					 * @param p Aktuln uzel stromu.
					 */
					void InternalClear(const Node* p);

					/**
					 * Rekurzvn metoda pro potn uzl ve stromu.
					 * @param p Aktuln uzel stromu.
					 * @return Funkce vrac poet uzl ve stromu s koenem p.
					 */
					int InternalCount(const Node* p) const;

					/**
					 * Rekurzvn metoda pro prchod stromem zpsobem in order.
					 * @param p Aktuln uzel stromu.
					 */
					void InternalInorderTraversal(Node* p);

					/**
					 * Rekurzvn metoda pro prchod stromem zpsobem pre order.
					 * @param p Aktuln uzel stromu.
					 */
					void InternalPreorderTraversal(Node* p);

					/**
					 * Rekurzvn metoda pro prchod stromem zpsobem post order.
					 * @param p Aktuln uzel stromu.
					 */
					void InternalPostorderTraversal(Node* p);

					/**
					 * Rekurzvn metoda pro vypisovn intern struktury stromu.
					 * @param p Aktuln uzel stromu.
					 * @param Indentation Odsazen od levho okraje, poet tabultor.
					 */
					void InternalReportStructure(Node* p, const int Indentation) const;

					/**
					 * Metoda demonstruje akci provdnou s uzlem pi prchodu stromem. V tomto ppad vypisuje kl na standardn vstup.
					 * @param p Aktuln uzel stromu
					 */
					void DoSomeAction(Node* p) const;

					/**
					 * Ukazatel na koen stromu.
					 */
					Node* mRoot;
				};


				inline BinarySearchTree::BinarySearchTree()
				{
					mRoot = NULL;
				}

				inline BinarySearchTree::~BinarySearchTree()
				{
					Clear();
				}

				inline void BinarySearchTree::Insert(const string Key)
				{
					InternalInsert(mRoot, Key);
				}

				inline void BinarySearchTree::Delete(const string Key)
				{
					InternalDelete(mRoot, Key);
				}

				inline bool BinarySearchTree::Search(const string Key) const
				{
					return InternalSearch(mRoot, Key);
				}

				inline void BinarySearchTree::Clear()
				{
					InternalClear(mRoot);
					mRoot = NULL;
				}

				inline bool BinarySearchTree::IsEmpty() const
				{
					return mRoot == NULL;
				}

				inline int BinarySearchTree::Count() const
				{
					return InternalCount(mRoot);
				}

				inline void BinarySearchTree::InorderTraversal()
				{
					InternalInorderTraversal(mRoot);
				}

				inline void BinarySearchTree::PreorderTraversal()
				{
					InternalPreorderTraversal(mRoot);
				}

				inline void BinarySearchTree::PostorderTraversal()
				{
					InternalPostorderTraversal(mRoot);
				}

				inline void BinarySearchTree::ReportStructure() const
				{
					InternalReportStructure(mRoot, 0);
				}
			}
		}
	}
}
