#pragma once

#include "BinarySearchTree.h"

namespace ADSLibrary
{
	namespace DataStructures
	{
		namespace LinkedStructures
		{
			namespace OOP
			{
				/**
	 * Abstraktn tda reprezentujc itertor v binrnm stromu.
	 * Itertor je zaloen na vyuit zsobnku, kter de-facto simuluje rekurzvn prchod stromem Inorder.
	 * Na rozdl od rekurze jsme ale v tomto ppad vykonvn prchodu stromem schopni
	 * "pozdret", vrtit data z aktulnho uzlu a potom pokraovat dle v prchodu stromem.
	 *
	 * @author      Ji Dvorsk <jiri.dvorsky@vsb.cz>
	 */
				class BinaryTreeIteratorBase
				{
				public:
					/**
					 * Konstruktor.
					 * @param Tree Strom, pes kter se bude iterovat
					 */
					BinaryTreeIteratorBase(BinarySearchTree* Tree);

					/**
					 * Reset itertoru. Potomci, pekrvajc tuto metodu mus odpovdajcm zpsobem inicializovat itertor (naplnit zsobnk uzly stromu)
					 */
					virtual void Reset() = 0;

					/**
					 * Test ukonen innosti itertoru. Tento test je pro vechny itertory shodn.
					 * @return Funkce vrac true pokud je iterace stromem ukonena, jinak false.
					 */
					bool IsEnd() const;

					/**
					 * Posun na dal uzel stromu ve smru iterace.
					 */
					virtual void MoveNext() = 0;

					/**
					 * Vrcen kle z aktulnho uzlu ve stromu.
					 * @return Funkce vrac kl z aktulnho uzlu ve stromu.
					 */
					string Current() const;

				protected:
					/**
					 * Jednoduch implementace zsobnku v poli pevn dlky. Vznam jednotlivch metod je popsn v pkladu specializovanm na implementaci zsobnku.
					 */
					class Stack
					{
					public:
						Stack();
						void Clear();
						void Push(BinarySearchTree::Node* n);
						BinarySearchTree::Node* Pop();
						BinarySearchTree::Node* Top() const;
						bool IsEmpty() const;
					private:
						static const int StackSize = 100;
						BinarySearchTree::Node* mData[StackSize];
						int mSP;
					};

					/**
					 * Ukazatel na strom pes kter iterujeme.
					 */
					BinarySearchTree* mTree;

					/**
					 * Zsobnk, kam jsou vkldny rozpracovan uzly stromu.
					 * Na vrcholu zsobnku je vdy uzel stromu, kter je pipraven na zpracovn tj. vrcen uivateli prostednictvm metody Current. Po proveden metody MoveNext se uzly
					 * v zsobnku peskldaj tak, aby opt byl na vrcholu dal prvek pipraven na zpracovn.
					 * Na konci iterace je zsobnk przdn.
					 */
					Stack mStack;
				};


				/**
				 * Tda implementujc itertor v binrnm stromu.
				 * Iterace probh podle uspodn prvk ve stromu, jinak eeno prvky jsou iterovny od nejmenho prvku ve stromu k nejvtmu.
				 *
				 * @author      Ji Dvorsk <jiri.dvorsky@vsb.cz>
				 */
				class BinaryTreeIterator : public BinaryTreeIteratorBase
				{
				public:
					/**
					 * Konstruktor.
					 * @param Tree Strom, pes kter se bude iterovat
					 */
					BinaryTreeIterator(BinarySearchTree* Tree);

					/**
					 * Reset itertoru. Do zsobnku jsou naskldny uzly lec na cest od koene k nejlevjmu uzlu.
					 */
					virtual void Reset();

					/**
					 * Posun na dal uzel stromu ve smru iterace. V tomto ppad si vyzvedneme ji zpracovan uzel na vrcholu zsobnku a hledme nejlevj uzel v jeho pravm podstromu.
					 * Jinak eeno: v pravm podstromu jsou vechny prvky vt ne ve vyzvednutm uzlu a pokud vezmeme jeho nejlevjho potomka, dostaneme se na uzel obsahujc nejbli vt prvek.
					 */
					virtual void MoveNext();
				};


				/**
				 * Tda implementujc itertor v binrnm stromu.
				 * Iterace probh proti smru uspodn prvk ve stromu, jinak eeno prvky jsou iterovny od nejvtho prvku ve stromu k nejmenmu.
				 * Itertor pracuje zrcadlov vi itertoru BinaryTreeIterator.
				 *
				 * @author      Ji Dvorsk <jiri.dvorsky@vsb.cz>
				 */
				class BinaryTreeReverseIterator : public BinaryTreeIteratorBase
				{
				public:
					/**
					 * Konstruktor.
					 * @param Tree Strom, pes kter se bude iterovat
					 */
					BinaryTreeReverseIterator(BinarySearchTree* Tree);

					/**
					 * Reset itertoru. Do zsobnku jsou naskldny uzly lec na cest od koene k nejpravjmu uzlu.
					 */
					virtual void Reset();

					/**
					 * Posun na dal uzel stromu ve smru iterace. V tomto ppad si vyzvedneme ji zpracovan uzel na vrcholu zsobnku a hledme nejpravj uzel v jeho levm podstromu.
					 * Jinak eeno: v levm podstromu jsou vechny prvky men ne ve vyzvednutm uzlu a pokud vezmeme jeho nejpravjho potomka, dostaneme se na uzel obsahujc nejbli men prvek.
					 */
					virtual void MoveNext();
				};


				inline BinaryTreeIteratorBase::BinaryTreeIteratorBase(BinarySearchTree* Tree)
				{
					mTree = Tree;
				}

				inline bool BinaryTreeIteratorBase::IsEnd() const
				{
					return mStack.IsEmpty();
				}

				inline string BinaryTreeIteratorBase::Current() const
				{
					return mStack.Top()->Key;
				}

				inline BinaryTreeIteratorBase::Stack::Stack()
				{
					mSP = 0;
				}

				inline void BinaryTreeIteratorBase::Stack::Clear()
				{
					mSP = 0;
				}

				inline void BinaryTreeIteratorBase::Stack::Push(BinarySearchTree::Node* n)
				{
					mData[mSP] = n;
					mSP += 1;
				}

				inline BinarySearchTree::Node* BinaryTreeIteratorBase::Stack::Pop()
				{
					mSP -= 1;
					return mData[mSP];
				}

				inline BinarySearchTree::Node* BinaryTreeIteratorBase::Stack::Top() const
				{
					return mData[mSP - 1];
				}

				inline bool BinaryTreeIteratorBase::Stack::IsEmpty() const
				{
					return mSP == 0;
				}

				inline BinaryTreeIterator::BinaryTreeIterator(BinarySearchTree* Tree)
					: BinaryTreeIteratorBase(Tree)
				{
				}

				inline BinaryTreeReverseIterator::BinaryTreeReverseIterator(BinarySearchTree* Tree)
					: BinaryTreeIteratorBase(Tree)
				{
				}

			}
		}
	}
}